"""
Model metadata module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.metadata import Metadata


class ModelMetadata(Metadata):
    """
    A class representing Model metadata.
    """


class ModelMetadataModel(ModelMetadata):
    """
    A class representing Model metadata.
    """
